import torch
import torch.nn as nn
from huggingface_hub import PyTorchModelHubMixin
from transformers import PreTrainedModel
from transformers.utils import PushToHubMixin

from audio_denoiser.modules.Permute import Permute
from audio_denoiser.modules.SimpleRoberta import SimpleRoberta
from audio_denoiser.modules.SpectrogramScaler import SpectrogramScaler


class AudioNoiseModel(nn.Module, PyTorchModelHubMixin):
    def __init__(self, config: dict):
        super(AudioNoiseModel, self).__init__()

        # Encoder layers
        self.config = config
        scaler_dict = config['scaler']
        self.scaler = SpectrogramScaler.from_dict(scaler_dict)
        self.in_channels = config.get('in_channels', 257)
        self.roberta_hidden_size = config.get('roberta_hidden_size', 768)
        self.model1 = nn.Sequential(
            nn.Conv1d(self.in_channels, 1024, kernel_size=1),
            nn.ELU(),
            nn.Conv1d(1024, 1024, kernel_size=1),
            nn.ELU(),
            nn.Conv1d(1024, self.in_channels, kernel_size=1),
        )
        self.model2 = nn.Sequential(
            Permute(0, 2, 1),
            nn.Linear(self.in_channels, self.roberta_hidden_size),
            SimpleRoberta(num_hidden_layers=5, hidden_size=self.roberta_hidden_size),
            nn.Linear(self.roberta_hidden_size, self.in_channels),
            Permute(0, 2, 1),
        )

    @property
    def sample_rate(self) -> int:
        return self.config.get('sample_rate', 16000)

    @property
    def n_fft(self) -> int:
        return self.config.get('n_fft', 512)

    @property
    def num_frames(self) -> int:
        return self.config.get('num_frames', 32)

    def forward(self, x, use_scaler: bool = False, out_scale: float = 1.0):
        if use_scaler:
            x = self.scaler(x)
        x1 = self.model1(x)
        x2 = self.model2(x)
        x = x1 + x2
        return x * out_scale



import os
from typing import List

import numpy as np
import torch
from torch import nn


class SpectrogramScaler(nn.Module):
    def __init__(self, mean: float, std: float):
        super().__init__()
        self.std = std
        self.mean = mean

    def to_dict(self):
        return dict(mean=self.mean, std=self.std)

    def __str__(self):
        return f'{self.__class__.__name__} {self.to_dict()}'

    @classmethod
    def from_dict(cls, d: dict):
        mean = d['mean']
        std = d['std']
        return SpectrogramScaler(mean, std)

    @staticmethod
    def train_scaler(audio_files: List[str], ambient_noise: List[torch.Tensor],
                     sample_rate=16000, n_fft=512) -> 'SpectrogramScaler':
        from data.AudioFileDataset import AudioFileDataset

        base_scaler = SpectrogramScaler(0, 1.0)
        rnd = np.random.RandomState(1)
        ds = AudioFileDataset(rnd, audio_files, ambient_noise, base_scaler, sample_rate=sample_rate, n_fft=n_fft)
        spectrogram_list = []
        for x, _ in ds:
            spectrogram_list.append(x)
        spectrogram_all = torch.cat(spectrogram_list)
        mean = torch.mean(spectrogram_all).item()
        std = torch.std(spectrogram_all).item()
        return SpectrogramScaler(mean, std)

    def forward(self, spectrogram: torch.Tensor) -> torch.Tensor:
        return (spectrogram - self.mean) / self.std



from typing import List, Any

import torch
from torch import nn


class View(nn.Module):
    def __init__(self, *shape: Any, contiguous: bool = False):
        super().__init__()
        self.contiguous = contiguous
        self.params = shape

    def forward(self, x: torch.Tensor):
        if self.contiguous:
            return x.contiguous().view(*self.params)
        else:
            return x.view(*self.params)



from typing import List

import torch.nn as nn


class Permute(nn.Module):
    def __init__(self, *dims: int):
        super(Permute, self).__init__()
        self.dims = dims

    def forward(self, x):
        return x.permute(*self.dims)



import math
from typing import Iterable
import torch
from torch import nn
import torch.nn.functional as F


def count_parameters(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def count_parameters_from_list(param_list: Iterable[nn.Parameter]):
    return sum(p.numel() for p in param_list)


def sin_pos_embeddings(seq_len: int, emb_dim: int) -> torch.Tensor:
    factors = torch.arange(1, emb_dim + 1)
    seq_range = torch.arange(0, seq_len)
    angles = math.pi * 2 * seq_range / seq_len
    return torch.sin(angles[:, None] * factors[None, :])


def concatenate_tensors(list_of_tensors):
    if isinstance(list_of_tensors[0], tuple):
        concatenated_tuples = []
        for i in range(len(list_of_tensors[0])):
            concatenated_tensors = torch.cat([t[i] for t in list_of_tensors], dim=0)
            concatenated_tuples.append(concatenated_tensors)
        return tuple(concatenated_tuples)
    else:
        return torch.cat(list_of_tensors, dim=0)


def batched_apply(model: nn.Module, x: torch.Tensor, batch_size: int = 64, detached: bool = False):
    output_list = []
    num_items = x.size(0)
    for b0 in range(0, num_items, batch_size):
        input_batch = x[b0:b0 + batch_size]
        output_batch = model(input_batch)
        if detached:
            if isinstance(output_batch, tuple):
                output_batch = tuple([x.detach() for x in output_batch])
            else:
                output_batch = output_batch.detach()
        output_list.append(output_batch)
    output = concatenate_tensors(output_list)
    return output


def sample_tensor(tensor: torch.Tensor, count: int):
    perm = torch.randperm(tensor.size(0), device=tensor.device)
    return tensor[perm[:count]]


def sample_tensors(tensor1: torch.Tensor, tensor2: torch.Tensor, count: int):
    if tensor2.size(0) != tensor1.size(0):
        raise ValueError('Different batch sizes!')
    perm = torch.randperm(tensor1.size(0), device=tensor1.device)
    selection = perm[:count]
    return tensor1[selection], tensor2[selection],


def fixed_hash(text: str):
    h = 0
    for ch in text:
        h = (h * 281 ^ ord(ch) * 997) & 0xFFFFFFFF
    return h


def tensor_hash(t: torch.Tensor):
    return fixed_hash(str(t.cpu().tolist()))


def unfold_2d(tensor: torch.Tensor, patch_size: int, step_size: int):
    # tensor shape: (channels, height, width)
    # result shape: (num_patches, channels, patch_size, patch_size)
    _, orig_height, orig_width = tensor.shape
    if orig_width < patch_size or orig_height < patch_size:
        raise ValueError(f'Tensor too small for patch_size={patch_size}')
    num_rows = math.ceil((orig_height - patch_size) / step_size) + 1
    num_cols = math.ceil((orig_width - patch_size) / step_size) + 1
    num_patches = num_rows * num_cols
    new_height = (num_rows - 1) * step_size + patch_size
    new_width = (num_cols - 1) * step_size + patch_size
    if new_height != orig_height or new_width != orig_width:
        padding_height = new_height - orig_height
        padding_width = new_width - orig_width
        tensor = F.pad(tensor, pad=(0, padding_width, 0, padding_height))
    # tensor shape: (channels, height, width)
    patches = tensor.unfold(1, patch_size, step_size)
    patches = patches.unfold(2, patch_size, step_size)
    patches = patches.contiguous().view((3, num_patches, patch_size, patch_size))
    patches_input = torch.permute(patches, (1, 0, 2, 3))
    return patches_input


def fold_2d(tensor: torch.Tensor, width: int, height: int):
    # tensor shape: (num_patches, channels, patch_size, patch_size)
    # result shape: (channels, height, width)
    num_patches, channels, patch_size, _ = tensor.shape
    step_size = patch_size
    num_rows = math.ceil(height / step_size)
    num_cols = math.ceil(width / step_size)
    if num_patches != num_rows * num_cols:
        raise ValueError(f'Expected num_rows={num_rows} * num_cols={num_cols} to be {num_patches}')
    new_height = num_rows * patch_size
    new_width = num_cols * patch_size
    tensor = torch.permute(tensor, (1, 0, 2, 3))
    # tensor: (channels, num_patches, patch_size, patch_size)
    tensor = tensor.view(channels, num_rows, num_cols, patch_size, patch_size)
    tensor = torch.permute(tensor, (0, 1, 3, 2, 4))
    # tensor: (channels, num_rows, patch_size, num_cols, patch_size)
    tensor = tensor.contiguous().view(channels, new_height, new_width)
    return torch.clone(tensor[:, :height, :width])



from typing import Optional, Union

import torch
import torchaudio.transforms as T


def create_spectrogram(waveform: torch.Tensor, n_fft: int, hop_length: int = None, power=2,
                       n_frames: Optional[int] = None, device: Union[torch.device, str] = None):
    if hop_length is None:
        hop_length = n_fft // 2
    spectrogram_transform = T.Spectrogram(n_fft=n_fft, win_length=n_fft, hop_length=hop_length, center=True,
                                          power=power)
    if device is not None:
        spectrogram_transform = spectrogram_transform.to(device)
    spectrogram = spectrogram_transform(waveform)
    if n_frames is not None:
        spectrogram = spectrogram[:, :, :n_frames]
    return spectrogram


def reconstruct_from_spectrogram(spectrogram: torch.Tensor, num_iterations=100,
                                 power=2, device: Union[torch.device, str] = None):
    _, half_fft, _ = spectrogram.shape
    n_fft = (half_fft - 1) * 2
    transform = T.GriffinLim(n_fft=n_fft, n_iter=num_iterations, rand_init=False, power=power)
    if device is not None:
        transform = transform.to(device)
    return transform(spectrogram)



import codecs

from setuptools import setup, find_packages

# Define project metadata
NAME = 'audio-denoiser'
VERSION = '0.1.2'
DESCRIPTION = 'A Python library for (speech) audio denoising.'
AUTHOR = 'Jose Solorzano'
URL = 'https://github.com/jose-solorzano/audio-denoiser'

REQUIRES = [
    'numpy>=1.23',
    'tqdm>=4.0.0',
    'torch>=2.1.0',
    'torchaudio>=2.1.0',
    'transformers>=4.28.0',
]

# Long description comes from README
with codecs.open('README.md', 'r', 'utf-8') as f:
    LONG_DESCRIPTION = f.read()

# Setup configuration
setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    url=URL,
    packages=['audio_denoiser', 'audio_denoiser.modules', 'audio_denoiser.helpers'],
    install_requires=REQUIRES,
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities',
    ],
    python_requires='>=3.9',
)



import torch
from torch import nn
from transformers import RobertaConfig, RobertaModel


class SimpleRoberta(nn.Module):
    def __init__(self, num_hidden_layers: int, hidden_size: int,
                 num_attention_heads: int = 8,
                 hidden_act: str = 'gelu'):
        super().__init__()
        config = RobertaConfig()
        config.vocab_size = 1
        config.pad_token_id = 0
        config.hidden_size = hidden_size
        config.num_hidden_layers = num_hidden_layers
        config.num_attention_heads = num_attention_heads
        config.hidden_act = hidden_act
        self.model = RobertaModel(config)

    def forward(self, inputs_embeds: torch.Tensor) -> torch.Tensor:
        rm_output = self.model(inputs_embeds=inputs_embeds)
        return rm_output.last_hidden_state



import logging
import math
from typing import Union, Optional
import torch
import torchaudio
from torch import nn
from audio_denoiser.helpers.torch_helper import batched_apply
from audio_denoiser.modules.AudioNoiseModel import AudioNoiseModel
from audio_denoiser.helpers.audio_helper import create_spectrogram, reconstruct_from_spectrogram

_expected_t_std = 0.23


class AudioDenoiser:
    def __init__(self, model_name='jose-h-solorzano/audio-denoiser-512-32-v1',
                 device: Union[str, torch.device] = None, num_iterations: int = 100):
        super().__init__()
        if device is None:
            is_cuda = torch.cuda.is_available()
            if not is_cuda:
                logging.warning('CUDA not available. Will use CPU.')
            device = torch.device('cuda:0') if is_cuda else torch.device('cpu')
        self.device = device
        self.model = AudioNoiseModel.from_pretrained(model_name).to(device)
        self.model.eval()
        self.model_sample_rate = self.model.sample_rate
        self.scaler = self.model.scaler
        self.n_fft = self.model.n_fft
        self.segment_num_frames = self.model.num_frames
        self.num_iterations = num_iterations

    @staticmethod
    def _sp_log(spectrogram: torch.Tensor, eps=0.01):
        return torch.log(spectrogram + eps)

    @staticmethod
    def _sp_exp(log_spectrogram: torch.Tensor, eps=0.01):
        return torch.clamp(torch.exp(log_spectrogram) - eps, min=0)

    @staticmethod
    def _trimmed_dev(waveform: torch.Tensor, q: float = 0.90) -> float:
        # Expected for training data is ~0.23
        abs_waveform = torch.abs(waveform)
        quantile_value = torch.quantile(abs_waveform, q).item()
        trimmed_values = waveform[abs_waveform >= quantile_value]
        return torch.std(trimmed_values).item()

    def process_waveform(self, waveform: torch.Tensor, sample_rate: int,
                         return_cpu_tensor: bool = False,
                         auto_scale: bool = False) -> torch.Tensor:
        """
        Denoises a waveform.
        @param waveform: A waveform tensor. Use torchaudio structure.
        @param sample_rate: The sample rate of the waveform in Hz.
        @param return_cpu_tensor: Whether the returned tensor must be a CPU tensor.
        @param auto_scale: Normalize the scale of the waveform before processing. Recommended for low-volume audio.
        @return: A denoised waveform.
        """
        waveform = waveform.to(self.device)
        if auto_scale:
            w_t_std = self._trimmed_dev(waveform)
            waveform = waveform * _expected_t_std / w_t_std
        if sample_rate != self.model_sample_rate:
            transform = torchaudio.transforms.Resample(orig_freq=sample_rate, new_freq=self.model_sample_rate)
            transform = transform.to(self.device)
            waveform = transform(waveform)
        hop_len = self.n_fft // 2
        spectrogram = create_spectrogram(waveform, n_fft=self.n_fft, hop_length=hop_len,
                                         device=self.device)
        num_a_channels = spectrogram.size(0)
        with torch.no_grad():
            results = []
            for c in range(num_a_channels):
                c_spectrogram = spectrogram[c]
                # c_spectrogram: (257, num_frames)
                fft_size, num_frames = c_spectrogram.shape
                num_segments = math.ceil(num_frames / self.segment_num_frames)
                adj_num_frames = num_segments * self.segment_num_frames
                if adj_num_frames > num_frames:
                    c_spectrogram = nn.functional.pad(c_spectrogram, (0, adj_num_frames - num_frames))
                c_spectrogram = c_spectrogram.view(fft_size, num_segments, self.segment_num_frames)
                # c_spectrogram: (257, num_segments, 32)
                c_spectrogram = torch.permute(c_spectrogram, (1, 0, 2))
                # c_spectrogram: (num_segments, 257, 32)
                log_c_spectrogram = self._sp_log(c_spectrogram)
                scaled_log_c_sp = self.scaler(log_c_spectrogram)
                pred_noise_log_sp = batched_apply(self.model, scaled_log_c_sp, detached=True)
                log_denoised_sp = log_c_spectrogram - pred_noise_log_sp
                denoised_sp = self._sp_exp(log_denoised_sp)
                # denoised_sp: (num_segments, 257, 32)
                denoised_sp = torch.permute(denoised_sp, (1, 0, 2))
                # denoised_sp: (257, num_segments, 32)
                denoised_sp = denoised_sp.contiguous().view(1, fft_size, adj_num_frames)
                # denoised_sp: (1, 257, adj_num_frames)
                denoised_sp = denoised_sp[:, :, :num_frames]
                denoised_waveform = reconstruct_from_spectrogram(denoised_sp,
                                                                 num_iterations=self.num_iterations,
                                                                 device=self.device)
                # denoised_waveform: (1, num_samples)
                results.append(denoised_waveform)
            cpu_results = torch.cat(results)
            return cpu_results.cpu() if return_cpu_tensor else cpu_results

    def process_audio_file(self, in_audio_file: str, out_audio_file: str, auto_scale: bool = False,
                           backend: Optional[str] = None):
        """
        Denoises an audio file.
        @param in_audio_file: An input audio file with a format supported by torchaudio.
        @param out_audio_file: Am output audio file with a format supported by torchaudio.
        @param auto_scale: Whether the input waveform scale should be normalized before processing. Recommended for low-volume audio.
        @param backend: The torchaudio I/O backend to use.
        """
        waveform, sample_rate = torchaudio.load(in_audio_file, backend=backend)
        denoised_waveform = self.process_waveform(waveform, sample_rate, return_cpu_tensor=True, auto_scale=auto_scale)
        torchaudio.save(out_audio_file, denoised_waveform, sample_rate=self.model_sample_rate)
